from __future__ import print_function
import numpy as np
from collections import Counter

class Candidate(object):
  def __init__(self, x, y, featurize, params):
    self.x,self.y = x,y
    self.features = featurize(x, y)
    self.score = self.score(params)

  def score(self, params):
    return sum(map(lambda ft: params[ft]*self.features[ft], self.features))

class SparseStructuredClassifier(object):
  def __init__(self, featurize):
    self.params = Counter({})
    self.featurize = featurize

  def nbest(self, x):
    """
      Returns: a list of the best candidates based on x and the current parameter
    """
    return []

  def learn(self, examples, step_size=1e-2, total_iters=6, verbose=False):
    """
    Train this classifier using SGD
    Inputs:
    - examples: list of dict keyed by x and y
    - step_size: step size of sgd
    Returns:
    - A history of losses for plotting and analysis
    """
    loss_history = []
    for t in range(total_iters):
      loss_t = 0
      for ex in examples:
        bests = self.nbest(ex['x'])
        loss, grad = self.loss(bests, ex['y'])
        loss_t += loss
        for feat in grad:
          self.params[feat] -= step_size * grad[feat]
      loss_history += [loss_t / len(examples)]
      if verbose and t % 1 == 0:
        print('iteration %d / %d: loss %f' % (t+1, total_iters, loss_history[-1]))
    return loss_history

  def predict(self, x):
    return self.nbest(x)[0].y

  def evaluate(self, examples):
    return np.mean([self.predict(ex['x'])==ex['y'] for ex in examples])


class MulticlassClassifier(SparseStructuredClassifier):
  def __init__(self, featurize, labels):
    super(MulticlassClassifier, self).__init__(featurize)
    self.labels = labels

  def loss(self, cands, ystar):
    """
    Inputs:
    - cands: a list of candidates that are featurized and scored
    - ystar: the correct answer
    Returns
    - tupple of (loss, gradient), where the gradient should be in a Counter
    """
    # TODO: implement the max margin loss
    return 0, {'':0}

  def nbest(self, x):
    cands = [*map(lambda y: Candidate(x, y, self.featurize, self.params), self.labels)]
    cands.sort(key=lambda c: c.score, reverse=True)
    return cands

class RatingsClassifier(SparseStructuredClassifier):
  """
  For ratings, we care about the difference, and not just correct/incorrect.
  """

  def loss(self, cands, ystar):
    """
    Inputs:
    - cands: a list of candidates that are featurized and scored
    - ystar: the correct answer
    returns the loss and gradient. The gradient should be in a Counter
    """
    ### TODO implement a margin loss that approximates the distance based evaluation function
    pass

  def evaluate(self, examples):
    return np.mean([abs(self.predict(ex['x'])-ex['y']) for ex in examples])
